#ifndef h_ocrinterfaces__
#define h_ocrinterfaces__

// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: iupr common header files
// File: ocrinterfaces.h
// Purpose: interfaces to OCR system components
// Responsible: tmb
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de


#include <stdlib.h>
#include "narray.h"
#include "narray-util.h"
#include "smartptr.h"
#include "misc.h"
#include "coords.h"
#include "nustring.h"

namespace colib {
    class idmap {
        narray< autoref<intarray> > ids;
        narray< autoref<intarray> > segments;
    public:
        idmap() {
            clear();
        }
        void segments_of_id(intarray &result, int id) {
            CHECK_ARG(id >= 0);
            if(id >= segments.length())
                result.clear();
            else
                copy(result,*segments(id));
        }
        void ids_of_segment(intarray &result,int segment) {
            CHECK_ARG(segment >= 0);
            if(segment >= ids.length())
                result.clear();
            else
                copy(result,*ids(segment));
        }
        void associate(int id, int segment) {
            // FIXME ids should start with 1 -> CHECK_ARG(id>0) ?
            CHECK_ARG(id>=0 && id<100000);
            CHECK_ARG(segment>=0 && segment<100000);
            while(!(segment<ids.length())) ids.push();
            while(!(id<segments.length())) segments.push();
            if(first_index_of(*ids(segment),id)<0)
                ids(segment)->push(id);
            if(first_index_of(*segments(id),segment)<0)
                segments(id)->push(segment);
        }
        void clear(int n=100000) {
            ids.dealloc();
            segments.dealloc();
        }
    };

    struct IComponent {
        virtual ~IComponent() {}
        virtual const char *description() = 0;
    };

    struct ICleanupGray : IComponent {
        virtual void cleanup(bytearray &out,bytearray &in) = 0;
    };

    struct ICleanupBinary : IComponent {
        virtual void cleanup(bytearray &out,bytearray &in) = 0;
    };

    struct ISegmentPage : IComponent {
        virtual void segment(intarray &out,bytearray &in) = 0;
    };

    struct ISegmentLine : IComponent {
        virtual void charseg(intarray &out,bytearray &in) = 0;
    };

    struct IBinarize : IComponent {
        virtual void binarize(bytearray &out,bytearray &in) {
            floatarray temp;
            copy(temp,in);
            binarize(out,temp);
        }
        virtual void binarize(bytearray &out,floatarray &in) = 0;
    };

    struct ICharLattice : virtual IComponent {
        // clear the language model
        virtual void clear() = 0;
        // get a single new state
        virtual int newState() = 0;
        // add a transition between the given states
        virtual void addTransition(int from,int to,nuchar output,float cost,int id) = 0;
        // set the start state
        virtual void setStart(int node) = 0;
        // set a state as an accept state
        virtual void setAccept(int node,float cost=0.0) = 0;
        // obtain codes for "specials" (language model dependent)
        virtual int special(const char *s) = 0;
        // compute the best path through the language model
        // useful for simple OCR tasks and for debugging
        virtual void bestpath(nustring &result) = 0;
        // destroy the language model
        virtual ~ICharLattice() {}
        // simple interface for line recognizers
        virtual void setString(nustring &text,floatarray &costs,intarray &ids) {
            int n = text.length();
            intarray states;
            states.clear();
            for(int i=0;i<n+1;i++)
                states.push(newState());
            for(int i=0;i<n;i++)
                addTransition(states[i],states[i+1],text[i],costs[i],ids[i]);
            setStart(states[0]);
            setAccept(states[n]);
        }
    };

    struct ICharacterClassifier : IComponent {
        // classify a character without any information about position on the line; this
        // may throw an exception if it's not implemented
        virtual void set(bytearray &input_image) = 0;
        // classify a character with information about position on the line; this may
        // throw an exception if it's not implemented
        virtual void set(bytearray &image,int base_y, int xheight_y, int descender_y, int ascender_y) = 0;
        // number of classes returned
        virtual int length() = 0;
        // UTF-8 encoded character or character string; note that some classifiers may return multiple characters
        // per class
        virtual void cls(nustring &result, int i) = 0;
        // cost value for this classification; lower costs = better
        virtual float cost(int i) = 0;
        // convenience function for getting the best output (useful for debugging)
        virtual void best(nustring &result) {
            int mi = -1;
            float mc = 1e30;
            for(int i=0;i<length();i++) {
                if(cost(i)<mc) {
                    mi = i;
                    mc = cost(i);
                }
            }
            if(mi>=0)
                cls(result, mi);
            else
                result.clear();
        }
        // destructor
        virtual ~ICharacterClassifier() {}
    };

    struct IRecognizeLine : IComponent {
        // recognize a text line
        virtual void recognizeLine(ICharLattice &result,colib::idmap &components,intarray &segmentation,bytearray &image) = 0;
        // destructor
        virtual ~IRecognizeLine() {}
    };

    struct ITrainChars : IComponent {
        // train a character (commonly, this only stores data in the model; training is via an external program)
        // this may be also train on ligatures (if supported),
        // that's why `characters' is a nustring.
        virtual void train(bytearray &input_image,nustring &characters,const char *context=0) = 0;
        // train a character (commonly, this only stores data in the model; training is via an external program)
        virtual void train(bytearray &image,int base_y, int xheight_y, int descender_y,
                           int ascender_y,nustring &characters,const char *context=0) = 0;
        // destructor
        virtual ~ITrainChars() {}
    };

    struct ITrainLines : IComponent {
        // train based on an unsegmented line
        virtual void trainLine(nustring &chars,bytearray &image,const char *context=0) = 0;
        // destructor
        virtual ~ITrainLines() {}
    };

    struct IBestPath : virtual IComponent {
        virtual void bestpath(nustring &result,floatarray &costs,intarray &ids,intarray &states) = 0;
    };

}

#endif
