// Copyright 2006-2007 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-adaptive
// File: decorations.cc
// Purpose: rule-based detection of paragraphs and headers
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de


#include <stdlib.h>
#include <stdio.h>
#include "layout-rules.h"
#include "ocr-utils.h"

using namespace ocropus;
using namespace colib;

namespace {

    static void get_bigger_than_flags(narray<bool> &bigger_than_prev,
                                      narray<bool> &bigger_than_next,
                                      narray<rectangle> &line_boxes) {
        makelike(bigger_than_prev, line_boxes);
        makelike(bigger_than_next, line_boxes);
        for(int i = 0; i < line_boxes.length(); i++) {
            int h = line_boxes[i].height();
            bigger_than_prev[i] = i == 0 || h > line_boxes[i-1].height() * 1.5;
            bigger_than_next[i] = i + 1 == line_boxes.length() 
                               || h > line_boxes[i+1].height() * 1.5;
        }
    }

    static void preinit_header_flags_by_medians(narray<bool> &result,
                                                intarray &line_medians) {
        makelike(result, line_medians);
        for(int i = 0; i < result.length(); i++) {
            result[i] = result.length() > 1 
                        && (
                            (i + 1 <  result.length() 
                            &&i > 0 
                            &&line_medians[i] >= 1.4 * line_medians[i + 1]  
                            &&line_medians[i] >= 1.1 * line_medians[i - 1]
                            )

                            || i == 0
                            && line_medians[i] >= 1.5 * line_medians[i + 1]

                            || (i + 2 < result.length()
                            && line_medians[i] >= 1.5 * line_medians[i + 2])
                
                            || ((i + 3) < result.length() 
                            && line_medians[i] >= 1.6 * line_medians[i + 3])

                            || ((i + 4) < result.length() 
                            && line_medians[i] >= 1.7 * line_medians[i + 4] )
                           );
        }
    }


    // I think this is by Ambrish --IM
    void runlength_histogram(narray<float> &hist, rectangle line_box, bytearray &img){
        fill(hist,0);
        int runlength = 0;
        int flag = 0;
        for(int j = line_box.y0; j < line_box.y1; j++){       
            for(int k = line_box.x0; k < line_box.x1; k++) {
                if (img(k,j) == 0) {
                    runlength++;
                    flag = 1;
                }
                if (img(k,j) != 0 && flag == 1) {
                    flag = 0;
                    if(runlength < hist.length())
                        hist(runlength)++;
                    runlength = 0;
                }
            }
        }    
    }   

    void runlength_histogram(narray<float> &hist, bytearray &img) {
        CHECK_ARG(background_seems_white(img));
        runlength_histogram(hist, rectangle(0, 0, img.dim(0), img.dim(1)), img);
    }

    void runlength_histogram(narray<float> &hist, ILines &lines, int line) {
        bytearray gray, mask;
        lines.line(gray, mask, line); // the mask should be the real binarization...
        make_background_white(mask);
        runlength_histogram(hist, mask);
    }

    int find_median(narray<float> &hist){
        int index= 0;
        float partial_sum = 0, sum = 0 ;
        for(int i = 0; i < hist.length(); i++) sum += hist(i);
        for(int j = 0; j < hist.length(); j++) hist(j) /= sum;
        while(partial_sum < 0.5) {
            partial_sum += hist(index);
            index++;
        }
        return index;
    }

    void detect_sections(narray<bool> &result, narray<rectangle> &line_boxes, int j) {
        makelike(result, line_boxes);
        if (!line_boxes.length())
            return;

        narray<bool> bigger_than_prev;
        narray<bool> bigger_than_next;
        get_bigger_than_flags(bigger_than_prev, bigger_than_next, line_boxes);

        intarray skips;
        makelike(skips, line_boxes);
        skips[0] = 30000;
        for(int i = 1; i < result.length(); i++) {
            int vskip = line_boxes[i-1].y0 - line_boxes[i].y1;
            skips[i] = vskip;
        }

        const double afterskip_coef = .7;  
        result[j] = result.length() > 1 
                    &&(j + 1 < result.length() 
                    && skips[j] > 1.5 * line_boxes[j].height() 
                    && skips[j + 1] > afterskip_coef * line_boxes[j].height()
                  
                    || j + 2 < result.length() 
                    && skips[j] > 1.5 * line_boxes[j].height()
                    && skips[j + 2] > afterskip_coef * line_boxes[j+1].height()
                  
                    || j > 0 
                    && j + 1 < result.length()
                    && result[j-1]
                    && skips[j + 1] > afterskip_coef * line_boxes[j].height()
            );   
    }

    // well, sorry for pointers - explained below
    void detect_headlines_internal(narray<bool> &result, narray<rectangle> &line_boxes, bytearray *image, ILines *lines) {
        int n = line_boxes.length();
        if (!n)
            return;
        intarray line_medians(n);
        int median = 0;
        floatarray hist(100), page_hist(100);
        fill(hist,0);
        for(int i = 0; i < n; i++) {
            // here's why we need pointers: we can have either the image or ILines here
            if(image)
                runlength_histogram(hist, line_boxes[i], *image);
            else if(lines)
                runlength_histogram(hist, *lines, i);
            else
                throw "detect_headlines> need either image or lines";

            add(page_hist, hist);
            line_medians[i] = find_median(hist);
        }
        median = find_median(page_hist);
        
        preinit_header_flags_by_medians(result, line_medians);

        for(int k= 0; k<result.length(); k++) {
            if (result[k] && (line_medians[k] <= median + 1)){
                detect_sections(result, line_boxes, k);
            }    
        }        
    }

}


namespace ocropus {


    void detect_headlines(narray<bool> &result, narray<rectangle> &line_boxes, bytearray &image) {
        detect_headlines_internal(result, line_boxes, &image, NULL);
    }

    void detect_headlines(narray<bool> &result, ILines &lines) {
        narray<rectangle> line_boxes(lines.linesCount());
        for(int i = 0; i < line_boxes.length(); i++)
            line_boxes[i] = lines.bbox(i);
        detect_headlines_internal(result, line_boxes, NULL, &lines);
    }

    void detect_paragraphs(narray<bool> &result, narray<rectangle> &line_boxes) {
        makelike(result, line_boxes);
        if (!line_boxes.length())
            return;

        for(int i = 0; i < result.length() - 1; i++) {
            rectangle &rect = line_boxes[i];
            rectangle &next = line_boxes[i + 1];

            // A line is a paragraph starter if, compared to the next line:
            // it's above
            // its right edge is more or less the same
            // its left edge is shifted to the right
            result[i] = next.y0 < rect.y0
                     && abs(next.x1 - rect.x1) < 0.5 * rect.height()
                     && rect.x0 - next.x0 > 0.5 * rect.height();
        }
        // The last line is never a paragraph starter.
        if (result.length())
            result[result.length() - 1] = false;
    }

    void detect_paragraphs(narray<bool> &result, ILines &lines) {
        narray<rectangle> line_boxes(lines.linesCount());
        for(int i = 0; i < line_boxes.length(); i++)
            line_boxes[i] = lines.bbox(i);
        detect_paragraphs(result, line_boxes);
    }

    void detect_paragraphs(RecognizedPage &result, ILines &lines) {
        narray<bool> a;
        detect_paragraphs(a, lines);
        result.setParagraphs(a);
    }

    void detect_headlines(RecognizedPage &result, ILines &lines) {
        narray<bool> a;
        detect_headlines(a, lines);
        result.setHeaders(a);
    }
};
