// Copyright 2007 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: 
// File: ocr-recognize-line-simple.cc
// Purpose: a simple implementation of ILineOCR without rescaling support
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de, www.ocropus.org

#include <stdarg.h>
#include "ocr-recognize-line.h"
#include "imglib.h"
#include "imgops.h"
#include "ocr-utils.h"
#include "line-info.h"
#include "ocr-segmentations.h"

using namespace colib;
using namespace imglib;
using namespace ocropus;

namespace {
    void binarize_masked(bytearray &result,
                         bytearray &gray,
                         bytearray &mask) {
        CHECK_ARG(samedims(gray, mask));
        CHECK_ARG(background_seems_white(gray));
        CHECK_ARG(background_seems_black(mask));

        bytearray dilated_mask;
        copy(dilated_mask, mask);
        dilate_circle(dilated_mask, 2);
        makelike(result, mask);
                                            
        // FIXME: determine the threshold in a more clever way
        int threshold = 128;
        for(int i = 0; i < result.length1d(); i++) {
            if (dilated_mask.at1d(i))
                result.at1d(i) = gray.at1d(i) > threshold ? 255 : 0;
            else
                result.at1d(i) = 255;
        }
    }


    void segment_masked(intarray &result,
                        ISegmentLine &segmenter,
                        bytearray &gray,
                        bytearray &mask) {

        bytearray binarized;
        binarize_masked(binarized, gray, mask);

        get_ocr_timings().line_segmenter.start();
        segmenter.charseg(result, binarized);
        get_ocr_timings().line_segmenter.stop();
        
        check_line_segmentation(result);
        ASSERT(samedims(result, binarized));
    }

    void train_on_segmentation(ITrainChars &train_chars,
                               intarray &orig_cseg,
                               bytearray &image,
                               nustring &text,
                               floatarray &costs,
                               float adaption_threshold) {
        
        CHECK_ARG(samedims(orig_cseg, image));
        CHECK_ARG(samedims(text, costs));
        
        intarray cseg;
        copy(cseg, orig_cseg);
        make_line_segmentation_black(cseg);
        set_line_number(cseg, 0);
        
        rectarray bboxes;
        bounding_boxes(bboxes, cseg);

        bytearray line;
        nustring ground_truth;
        ground_truth.resize(1);
    
        float intercept;
        float slope;
        float xheight;
        float ascender_rise;
        float descender_sink;
        if (!get_extended_line_info(intercept, slope, xheight,
                                    descender_sink, ascender_rise, cseg)) {
            return;
        }

        bytearray char_image;
        
        for (int i = 1; i < bboxes.length(); i++) {
            // we should silently skip segmentation gaps
            if(bboxes[i].empty())
                continue;

            if(costs[i - 1] > adaption_threshold)
                continue;

            //crop_masked(char_image, image, bboxes[i], mask, 255);
            //pad_by(char_image, 3, 3, byte(255));
            //intercept += 3 - bboxes[i].y0;
            //XXX: dilate?
            //log("preved");
            copy(char_image, image);
            for(int k = 0; k < char_image.length1d(); k++) {
                if(cseg.at1d(k) != i)
                    char_image.at1d(k) = 255;
                else
                    char_image.at1d(k) = image.at1d(k);
            }

            int center_x = (bboxes[i].x0 + bboxes[i].x1) / 2;
            float baseline = intercept + center_x * slope ;
            
            ground_truth[0] = text[i - 1];
            
            train_chars.train(char_image,
                              int(baseline + .5),
                              int(baseline + xheight + .5),
                              int(baseline - descender_sink + .5),
                              int(baseline + xheight + ascender_rise + .5),
                              ground_truth);
        }
    }


    enum LineOcrState {
        JUST_STARTED,
        HAVE_IMAGE,
        REJECT,
        HAVE_IDMAP,
        HAVE_TRAIN_DATA
    };
    
    ///////////////////////////////////////////////////////////////////////////

    struct SimpleLineOCR : public ILineOCR {
        autodel<IRecognizeLine> recognize_line;
        autodel<ISegmentLine> segment_line;
        autodel<ITrainChars> train_chars;
        
        char description[1000];
        LineOcrState state;

        bytearray gray;
        intarray overseg;
        intarray true_seg;
        intarray true_ids;

        idmap id_map;
        
        // _________________________   bookkeeping   ___________________________
    
        SimpleLineOCR() : state(JUST_STARTED) {}
        
        virtual void inject(IRecognizeLine *ptr) {
            if(!!recognize_line) throw "IRecognizeLine already injected";
            recognize_line = ptr;
        }
        
        virtual IRecognizeLine *take_RecognizeLine() {
            return recognize_line.move();
        }

        virtual void inject(ISegmentLine *ptr) {
            if(!!segment_line) throw "ISegmentLine already injected";
            segment_line = ptr;
        }
        
        virtual ISegmentLine *take_SegmentLine() {
            return segment_line.move();
        }
        
        virtual void inject(ITrainChars *ptr) {
            if(!!train_chars) throw "ITrainChars already injected";
            train_chars = ptr;
        }

        virtual ITrainChars *take_TrainChars() {
            return train_chars.move();
        }
        
        /// Give a description of the line (for debugging purposes).
        virtual void setDescription(const char *format, ...) {
            va_list va;
            va_start(va, format);
            vsnprintf(description, sizeof(description), format, va);
            va_end(va);
        }

        // _________________________   recognition   ___________________________
        
        /// Give a grayscale image with a mask to the recognizer.
        virtual void set(colib::bytearray &in_gray,
                         colib::bytearray &mask) {
            if(!segment_line)
                throw "no ISegmentLine injected";
            if(!recognize_line)
                throw "no IRecognizeLine injected";

            if(in_gray.rank() != 2 || in_gray.total == 0) {
                overseg.clear();
                true_seg.clear();
                state = REJECT;
                return;
            }
            CHECK_ARG(samedims(in_gray, mask));
            CHECK_ARG(background_seems_black(mask));
            copy(gray, in_gray);
            make_background_white(gray);
            //if(rejectLineIfNeeded()) return;
            segment_masked(overseg, *segment_line, gray, mask);
            state = HAVE_IMAGE;
        }

        /// Get the oversegmentation of the original image.
        /// Might have gaps.
        virtual void getOversegmentation(colib::intarray &result) {
            copy(result, overseg);
        }
         
        /// Recognize and fill the char lattice.
        /// Note that idmap stuff is by design encapsulated inside ILineOCR. 
        virtual void recognize(colib::ICharLattice &lattice) {
            if(state == REJECT) return;
            CHECK_ARG(state == HAVE_IMAGE || state == HAVE_IDMAP);
            id_map.clear();
            recognize_line->recognizeLine(lattice, id_map, overseg, gray);
            state = HAVE_IDMAP;
        }

        // _______________________   adaptation   ______________________________

        void get_true_seg_if_needed() {
            if(state == HAVE_TRAIN_DATA)
                return;
            CHECK_ARG(state == HAVE_IDMAP);
            ocr_result_to_charseg(true_seg, id_map, true_ids, overseg, /*map_all:*/ false); // 'true' just doesn't work
            state = HAVE_TRAIN_DATA;
        }

        /// Given a result of the best path search,
        /// train the given ITrainChars instance.
        /// Note that it's better to train with train() then
        /// through getSegmentation() since train() might use
        /// rescaled or otherwise transformed characters
        /// better suited for adaptation.
        virtual void adapt(colib::nustring &text,
                           colib::floatarray &costs,
                           colib::intarray &ids,
                           float adaption_threshold) {
            if(state == REJECT) return;
            copy(true_ids, ids);
            if(!train_chars) return;
            get_true_seg_if_needed();
            CHECK_ARG(state == HAVE_TRAIN_DATA);
            train_on_segmentation(*train_chars, true_seg, gray, text,
                                   costs, adaption_threshold);
        }

        /// Return "true" segmentation of the original image.
        /// Might have gaps.
        virtual void getSegmentation(colib::intarray &result) {
            get_true_seg_if_needed();
            copy(result, true_seg);
        }       
    };
}

namespace ocropus {
    ILineOCR *make_SimpleLineOCR() {
        return new SimpleLineOCR();
    }
}
