// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-adaptive
// File: ocr-recognize-page.h
// Purpose: interface to the page reader
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de


#ifndef h_ocr_recognize_page_
#define h_ocr_recognize_page_

#include "colib.h"
#include "lines.h"
#include "ocr-recognize-line.h"

namespace ocropus {

    /// A RecognizedPage is a transport object responsible for keeping the OCR results.
    /// Also, it might communicate some options to the OCR to select which data needs
    /// to be in the OCR output.
    class RecognizedPage {
        colib::narray<colib::nustring>   line_texts;
        colib::narray<colib::floatarray> line_costs;
        colib::rectarray line_bboxes;
        colib::strbuf desc;
        colib::narray<bool> is_header;
        colib::narray<bool> is_paragraph;
        int w, h;
        colib::strbuf time_report;
    public:
        int width() {
            return w;
        }
        int height() {
            return h;
        }
        bool isParagraph(int index) {
            return is_paragraph(index);
        }
        bool isHeader(int index) {
            return is_header(index);
        }
        void setParagraphs(colib::narray<bool> &n) {
            colib::copy(is_paragraph, n);
        }
        void setHeaders(colib::narray<bool> &n) {
            colib::copy(is_header, n);
        }
        void setWidth(int width) {
            w = width;
        }
        void setHeight(int height) {
            h = height;
        }
        const char *description() {
            return desc;
        }
        void setDescription(const char *s) {
            desc = s;
        }
        const char *timeReport() {
            return time_report;
        }
        void setTimeReport(const char *s) {
            time_report = s;
        }
        void setLinesCount(int n) {
            line_texts.resize(n);
            line_costs.resize(n);
            line_bboxes.resize(n);
            is_header.resize(n);
            is_paragraph.resize(n);
            fill(is_header, false);
            fill(is_paragraph, false);
        }
        int linesCount() {
            return line_texts.length();
        }
        void setText(colib::nustring &s, int index) {
            colib::copy(line_texts[index], s);
        }
        void text(colib::nustring &s, int index) {
            colib::copy(s, line_texts[index]);
        }
        void setCosts(colib::floatarray &c, int index) {
            colib::copy(line_costs[index], c);
        }
        void costs(colib::floatarray &c, int index) {
            colib::copy(c, line_costs[index]);
        }
        colib::rectangle bbox(int index) {
            return line_bboxes[index];
        }
        void setBbox(colib::rectangle bbox, int index) {
            line_bboxes[index] = bbox;
        }
    };
    
    // _________________________  recognize() functions  ______________________
    
    /// 1-pass recognition.
    /// Whether it adapts something or not is determined by the ILineOCR.
    /// Set pageno to -1 to avoid calling processPage() at all.
    void recognize(RecognizedPage &result,
                   ILines &lines,
                   ILineOCR &ocr,
                   colib::ICharLattice &lattice,
                   colib::IBestPath &bestpath,
                   float adaption_threshold,
                   int pageno = 0);
    
    /// 1-pass recognition.
    void recognize(RecognizedPage &result,
                   ILines &lines,
                   colib::ISegmentLine &lineseg,
                   colib::IRecognizeLine &rec,
                   colib::ICharLattice &lattice,
                   colib::IBestPath &bestpath,
                   float adaption_threshold,
                   int pageno = 0);

    /// 1-pass recognition with adaptation.
    void recognize(RecognizedPage &result,
                   ILines &lines,
                   colib::ISegmentLine &lineseg,
                   colib::IRecognizeLine &rec,
                   colib::ICharLattice &lattice,
                   colib::IBestPath &bestpath,
                   colib::ITrainChars &trainable,
                   float adaption_threshold,
                   int pageno = 0);

    /// 2-pass recognition.
    void recognize(RecognizedPage &result,
                   ocropus::ILines &lines,
                   colib::ISegmentLine &lineseg,
                   colib::IRecognizeLine &pass1,
                   colib::ICharLattice &lattice1,
                   colib::IBestPath &bestpath1,
                   colib::ITrainChars &trainable,
                   colib::IRecognizeLine &pass2,
                   colib::ICharLattice &lattice2,
                   colib::IBestPath &bestpath2,
                   float adaption_threshold,
                   int pageno = 0);

    void tesseract_recognize_blockwise(RecognizedPage &result,
                                       colib::bytearray &gray,
                                       colib::intarray &pageseg);
    
    void tesseract_recognize_blockwise(RecognizedPage &result,
                                       ILines &lines,
                                       int pageno = 0);
};

#endif
